// *** Inclusiones ***

#include <sys/types.h> // Tipos varios
#include <pthread.h> // Hebras
#include <sys/socket.h> // Socket
#include <netinet/in.h> // Red
#include <netdb.h> // Y ms red
#include <stdio.h> // Entrada y salida
#include <string.h> // Manejo de strings
#include <pwd.h> // getpwuid(), lol...


// *** Definiciones

#define PORTNUMBER 6699 // Puerto por omisin
#define BUFFSIZE 1024 // Tamao del buffer
#define NOMSIZE 16 // Tamao mximo del nombre de usuario
#define HOSTSIZE 64 // Tamao del string hostname


// *** hiloLectura()
// Hilo que recibe los mensajes desde el servidor y los muestra en pantalla
// (int*) void* arg: puntero a un socket

void* hiloLectura (void* arg) {
    
	// Puntero al socket
    int* socket;
	// Buffer de lectura
    char buf[BUFFSIZE];
    int n;
    
	// Asignamos el socket
    socket = arg;
    
	// Y nos dedicamos a leer desde el servidor y escribir en pantalla
    while ((n = recv(*socket, buf, sizeof(buf), 0)) > 0)
		write(1, buf, n);
    
    // Una vez que la conexin se termine, avisamos y cerramos
    printf("Conexin terminada.\n");
    close(*socket);
    exit(0);
}
    
    
// *** mcc ***
// My Chat Client
// mcc [host] [port]
// Cliente de chat

int main(int argc, char * argv[]) {
	
	// Socket
	int s;
    int n, len, port;
	// Buffer de escritura
    char buf[BUFFSIZE];
	// El resto de los chistes necesarios para conectarse
    char hostname[HOSTSIZE];
    struct hostent *hp;
    struct sockaddr_in name;
    
	// Tid de la hebra de lectura
    pthread_t myTid;
	// Chistes necesarios para obtener el nombre de usuario
	struct passwd* myPasswd;
	uid_t myUid;

	// Primero, analizamos los argumentos del proceso
    switch(argc) {
		case 1:
			// En caso de no tener argumentos, usamos el servidor propio y puerto por omisin
			if (gethostname(hostname, sizeof(hostname))) {
				perror("gethostname()");
				exit(1);
			}
			port = PORTNUMBER;
			break;
        
		case 2:
			// Puede que tengamos el puerto o el servidor como argumento
			// Diferenciamos primero del puerto
			
			// Si el puerto es un entero vlido
			port = atoi(argv[1]);
			if (port > 0) {
				// Usamos el servidor propio
				if (gethostname(hostname, sizeof(hostname))) {
					perror("gethostname()");
					exit(1);
				}
			}
			else {
				// En caso de tener el servidor como argumento
				strcpy(hostname, argv[1]);
				// Usamos el puerto por omisin
				port = PORTNUMBER;
			}
			break;
        
		case 3:
			// En caso de tener el servidor y puerto como argumentos
			// (Darse cuenta de que en otro orden nos fregamos)
			strcpy(hostname, argv[1]);
			port = atoi(argv[2]);
			break;
		
		default:
			// En caso de tener ms argumentos, mostramos el uso esperado
			printf("Usage: %s [server_name] [port]\n", argv[0]);
			exit(1);
    }
	
	// Para obtener el nombre de usuario:
	// Primero obtenemos nuestra uid
	myUid = getuid();
	// Luego obtenemos el nombre de usuario desde la estructura passwd
	myPasswd = getpwuid(myUid);
	if (myPasswd == NULL) {
		perror("getpwuid()");
		exit(1);
	}
	
	// Opcional: podemos obtener el nombre de usuario desde consola tambin (=
	/*
    printf("Escriba su nombre de usuario:");
    scanf("%s", buf);
	n = strlen(buf);
	*/
	
	// Guardamos el nombre de usuario en buf
	n = strlen(myPasswd->pw_name);
	memcpy(buf, myPasswd->pw_name, n+1);
	buf[n] = '\0';

	// Informamos en pantalla
	printf("Conectando a: %s, en puerto: %d. Usuario: %s.\n", hostname, port, buf);
    
    // Obtenemos los datos del servidor
    hp = gethostbyname(hostname);
	if(hp == NULL) {
		herror("gethostbyname()"); // herrar es equino...
		exit(1);
	}
    // Creamos el socket
    s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		perror("socket()");
		exit(1);
	}
    // Creamos la direccin del servidor
    name.sin_family = AF_INET;
    name.sin_port = htons(port);
    memcpy(&name.sin_addr, hp->h_addr_list[0], hp->h_length);
    len = sizeof(struct sockaddr_in);

    // Y nos conectamos al servidor
    if (connect(s, (struct sockaddr*) &name, len)) {
		perror("connect()");
		exit(1);
	}
    
	// Creamos un nuevo hilo para leer desde el servidor
    if (pthread_create (&myTid, NULL, hiloLectura, &s)) {
        perror("pthread_create()");
        exit(1);
    }
    // Se especifica que no se espera su valor de retorno (ahorro de recursos)
    if (pthread_detach(myTid)) {
        perror("pthread_detach()");
        exit(1);
    }
    
    // Y empezamos a enviar lo escrito desde entrada estndar hacia el servidor
    n = send(s, buf, strlen(buf)+1, 0);
    while ((n = read(0, buf, sizeof(buf))) > 0) {
        if (send(s, buf, n, 0) < 0) {
            perror("send");
            exit(1);
        }
    }
    
	// Una vez que terminamos de escribir, cerramos (si es que no nos ctr-c antes)...
    printf("Conexin terminada.\n");
    close(s);
    exit(0);
}