// *** Inclusiones ***

#include <sys/types.h> // Tipos varios
#include <pthread.h> // Hebras
#include <sys/socket.h> // Sockets
#include <netinet/in.h> // Conexin
#include <stdio.h> // Entrada y salida
#include <string.h> // Manejo de strings


// *** Definiciones ***

#define MAXUSUARIOS 4 // Nmero mximo de usuarios en el servidor
#define NOMSIZE 16 // Tamao mximo del nombre de usuario
#define BUFSIZE 1024 // Tamao del buffer de mensajes
#define PORTNUMBER 6699 // Puerto por omisin
#define T_ESPERA 5000 // Tiempo en microsegundos que duerme el hilo reenviante


// *** VARIABLES GLOBALES ***

int sockArr[MAXUSUARIOS+1]; // Arreglo de sockets
int usrNum; // Nmero de usuarios conectados
char buffRenv[BUFSIZE*MAXUSUARIOS*5]; // Buffer de reenvo. Pensado para un tope de cinco mensajes por usuario 
int buffNum; // ndice del buffer de reenvo
pthread_mutex_t sockMutex; // Mutex para acceder al arreglo de sockets
pthread_mutex_t buffMutex; // Mutex para acceder el buffer de reenvo



// *** hiloServidor() *** 
// En esta funcin corren los hilos que atienden a los usuarios

void* hiloServidor (void* arg) {
	
    // Socket es un puntero al socket asignado en el arreglo
    int* socket;
    // ndice apunta al final del nombre de usuario en el mensaje
    int indice;
    // Buffer que contiene el nombre de usuario ms el mensaje
	char mensaje[NOMSIZE+BUFSIZE+1];
    
    int n;
	
	// Asignamos a socket el puntero arg
    socket = arg;
    
    // Se obtiene el nombre de usuario.
    n = recv(*socket, mensaje, sizeof(mensaje), 0);
    // Si se recibe correctamente
    if (n > 0) {
        
        // Lo anunciamos al servidor
        printf("Usuario: %s. Socket: %d\n", mensaje, *socket);
       	
        // Lo anunciamos al resto de los usuarios
        // Entramos a la zona protegida del buffer
        pthread_mutex_lock(&buffMutex);
        // De no haber espacio en el buffer reclamamos
        if (buffNum+n+30 >=  sizeof(buffRenv))
            printf("Error: buffer lleno. Se descartar el mensaje para evitar inconsistencias.\n");
        // En caso contrario escribimos en el buffer
        else {
            memcpy(mensaje+n, " se ha conectado al servidor.\n", 30), 
            memcpy(buffRenv+buffNum, mensaje, n+30);
            buffNum += n+30;
            buffRenv[buffNum] = '\0';
        }
        // Y salimos de la zona protegida del buffer.
        pthread_mutex_unlock(&buffMutex);
        
        // El inicio de mensaje es "nombre_de_usuario: "
        mensaje[n-1] = ':';
        mensaje[n] = ' ';
        // Y recordamos el ndice
        indice = n+1;
    }
    
    // De lo contrario la hebra muere
    else {
        // Entramos a la zona protegida de los sockets 
        pthread_mutex_lock(&sockMutex);
        // Borramos el socket de la lista
        close(socket);
        *socket = 0;
        usrNum--;
        // Salimos de la zona protegida de los sockets
        pthread_mutex_unlock(&sockMutex);
        // Y fin
        pthread_exit(NULL);
    }
	
    // Por ltimo, el lazo de lectura
    // Leemos del socket, mientras existan posibles mensajes
	while ((n = recv(*socket, mensaje +indice, sizeof(mensaje), 0)) > 0) {
		
        // Podemos escribirlo al servidor, en caso de ser voyeuristas
        // write(1, mensaje, n+indice);
        
        // Copiamos el mensaje al buffer de reenvo
        // Entramos a la zona protegida del buffer
        pthread_mutex_lock(&buffMutex);
        // De no haber espacio en el buffer reclamamos
        if (buffNum+n+indice >=  sizeof(buffRenv))
            printf("Error: buffer lleno. Se descartar el mensaje para evitar inconsistencias.\n");
        // En caso contrario escribimos en el buffer
        else {
            memcpy(buffRenv+buffNum, mensaje, n+indice);
            buffNum += n+indice;
            buffRenv[buffNum] = '\0';
        }
        // Y salimos de la zona protegida del buffer.
        pthread_mutex_unlock(&buffMutex);
    }
    // Una vez que acaba la conexin, la hebra muere
	mensaje[indice-2] = '\0';
	// Lo anunciamos al servidor
	printf("Usuario %s se ha desconectado. Socket: %d\n", mensaje, *socket);
	
	// Lo anunciamos al resto de los usuarios
	// Entramos a la zona protegida del buffer
	pthread_mutex_lock(&buffMutex);
	// De no haber espacio en el buffer reclamamos
	if (buffNum+n+30 >=  sizeof(buffRenv))
		printf("Error: buffer lleno. Se descartar el mensaje para evitar inconsistencias.\n");
	// En caso contrario escribimos en el buffer
	else {
		memcpy(mensaje+indice-2, " se ha desconectado.\n", 21), 
		memcpy(buffRenv+buffNum, mensaje, indice+19);
		buffNum += indice+19;
		buffRenv[buffNum] = '\0';
	}
	// Y salimos de la zona protegida del buffer.
	pthread_mutex_unlock(&buffMutex);

    // Entramos a la zona protegida de los sockets 
    pthread_mutex_lock(&sockMutex);
    // Borramos el socket de la lista
    close(socket);
    *socket = 0;
    usrNum--;
    // Salimos de la zona protegida de los sockets
    pthread_mutex_unlock(&sockMutex);
    // Y fin
    pthread_exit(NULL);
}



// *** hiloReenvio()
// Reenva los mensajes a todos los usuarios

void* hiloReenvio (void* arg) {
    
    int i;
    
    // En un ciclo eterno
    while (1) {
     
        // Dormimos un rato
        usleep(T_ESPERA);
        // Luego entramos a la zona protegida del buffer
        pthread_mutex_lock(&buffMutex);
        
        // De haber mensajes pendientes      
        if (buffNum > 0) {
            
            // Entramos a la zona protegida de los sockets
            pthread_mutex_lock(&sockMutex);
            // Y enviamos los mensajes a todos los sockets vlidos
            for(i = 0; i < MAXUSUARIOS+1; i++)
                if (sockArr[i] > 0)
                    send (sockArr[i], buffRenv, buffNum, 0);
            // Salimos de la zona protegida de los sockets
            pthread_mutex_unlock(&sockMutex);
            // Reseteamos el ndice
            buffNum = 0;
        }
        // Y salimos de la zona protegida del buffer
        pthread_mutex_unlock(&buffMutex);
    }
}



// *** mcs ***
// My Chat Server
// mc [port]
// Servidor de chat

int main (int argc, char * argv[]) {

	int i, j, sockCont, port;
    
	pthread_t tid[MAXUSUARIOS+1], tidReenvio;
	
	char buf[10];
    int s, n, len;
    struct sockaddr_in name;
    
	// Primero, analizamos los argumentos del proceso
    switch(argc) {
		case 1:
			// En caso de no tener argumentos, usamos el puerto por omisin
			port = PORTNUMBER;
			break;
        
		case 2:
			// Si el puerto no es un entero vlido seguimos hasta el default
			port = atoi(argv[1]);
			if (port > 0)
				break;
		
		default:
			// En caso de tener ms argumentos, mostramos el uso esperado
			printf("Usage: %s [port]\n", argv[0]);
			exit(1);
    }
	
    // Inicializamos los mutex
    pthread_mutex_init(& sockMutex, NULL);
	pthread_mutex_init(& buffMutex, NULL);
    
    // Inicializamos los sockets en cero
    for (sockCont = 0; sockCont < MAXUSUARIOS+1; sockCont++)
        sockArr[sockCont] = 0;
    sockCont = 0;
    
	// Creamos el socket
    s = socket(AF_INET, SOCK_STREAM, 0);
	i = 1;
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*) &i, sizeof(i)) == -1) {
		perror("setsockopt()");
		exit(1);
	}
    // Creamos la direccin del servidor
    name.sin_family = AF_INET;
    name.sin_port = htons(port);
    name.sin_addr.s_addr = htonl(INADDR_ANY);
    len = sizeof(struct sockaddr_in);

    // Atamos la direccin al socket
   if (bind (s, (struct sockaddr*) &name, len)) {
		perror("bind");
		exit(1);
	}
    
    // Y empezamos. Reportamos el puerto de recepcin.
    printf("mcs en lnea. Escuchando requerimientos en el puerto %d.\n", port);
    
    // Iniciamos el hilo de reenvio
    if (pthread_create (&tidReenvio, NULL, hiloReenvio, NULL)) {
        perror("pthread_create()");
        exit(1);
    }
    // Se especifica que no se espera su valor de retorno (ahorro de recursos)
    if (pthread_detach(tidReenvio)) {
        perror("pthread_detach()");
        exit(1);
    }
    
    // Entramos al ciclo infinito de esperar clientes
	while (1) {
        
		//printf("Esperando por un cliente.\n");
		// Esperamos por una conexin
		if (listen(s, 5)) {
			perror("listen()");
			exit(1);
		}
		// Aceptamos la conexin en el socket libre
        //printf("Aceptando cliente.\n");
		sockArr[sockCont] = accept(s, (struct sockaddr*) &name, &len);
		if (sockArr[sockCont] < 0) {
			perror("accept()");
			exit(1);
		}
        // Entramos a la zona protegida de los sockets
        pthread_mutex_lock(&sockMutex);
        // En caso de tener cupos
        if (usrNum < MAXUSUARIOS) {
            
            printf("Recibida una conexin desde : %s. ",inet_ntoa(name.sin_addr));
            printf("Cantidad de usuarios: %d.\n", usrNum+1);
            
            // Se crea un nuevo hilo para ese usuario
            if (pthread_create (&tid[sockCont], NULL, hiloServidor, &sockArr[sockCont])) {
                perror("pthread_create()");
                exit(1);
            }
            // Se especifica que no se espera su valor de retorno (ahorro de recursos)
            if (pthread_detach(tid[sockCont])) {
                perror("pthread_detach()");
                exit(1);
            }
            // Lo consideramos en la lista
            usrNum++;
            // Y pasamos al siguiente socket libre
            do {
                sockCont++;
                sockCont %= MAXUSUARIOS+1;
            } while (sockArr[sockCont] != 0);
        }
        
        // En caso de estar lleno
        else {
            printf("Recibida una conexin desde : %s. ",inet_ntoa(name.sin_addr));
            printf("Servidor lleno (%d usuarios). Se negar el servicio.\n", usrNum+1);
            
            // Se le notifica al cliente
            send (sockArr[sockCont], "Error: servidor lleno.\n", 24, 0);
            // Y se cierra la conexin
            send (sockArr[sockCont], NULL, 0, 0);
            close (sockArr[sockCont]);
            sockArr[sockCont] = 0;
        }
        // Y salimos de la zona protegida de sockets
        pthread_mutex_unlock(&sockMutex);
	}
}
